use crate::util::embed::Embed;
use crate::util::user_util;
use crate::util::user_util::get_vc_id;
use crate::music::preview::Preview;
use lavalink_rs::client::LavalinkClient;
use lavalink_rs::model::search::SearchEngines;
use lavalink_rs::model::track::TrackLoadData;
use lavalink_rs::model::ChannelId;
use lavalink_rs::player_context::TrackInQueue;
use serenity::all::{Context, CreateEmbed, GuildId, Http, UserId};
use songbird::error::JoinError;
use std::sync::Arc;

/// Either queues the song, or start playing it instantly, depending on if there is already a song playing
pub async fn attempt_to_queue_song(
    ctx: &Context,
    llc: &LavalinkClient,
    guild_id: &GuildId,
    user_id: &UserId,
    username: &str,
    query: &str,
) -> CreateEmbed {
    if !user_util::is_user_connected_to_vc(ctx, guild_id, user_id).await {
        return Embed::create_error_respose(
            username,
            "You are not connected to a VC",
            "Connect to my vc to start controlling the music.",
        );
    }

    let connect_to = match get_vc_id(ctx, guild_id, user_id).await {
        Some(channel_id) => channel_id,
        None => {
            return Embed::create_error_respose(username, "Error", "Cannot find channel_id.");
        }
    };

    let songbird: &Arc<songbird::Songbird> = &songbird::get(ctx) // TODO match
        .await
        .expect("Cannot get Songbird.")
        .clone();

    let self_channel = user_util::get_self_vc_id(ctx, guild_id).await;
    if !user_util::is_self_connected_to_vc(ctx, &guild_id).await {
        // self is connected to vc, check if user is in same vc

        if self_channel.is_none() {
            // Connect to VC
            let handler = songbird.join_gateway(GuildId::new(guild_id.get()), connect_to).await;
            // 
            match handler {
                Ok((connection_info, _)) => {
                    let Ok(_) = llc
                    .create_player_context_with_data::<(ChannelId, std::sync::Arc<Http>)>(
                        guild_id.get(),
                        connection_info,
                        std::sync::Arc::new((
                            ChannelId::from(connect_to.get()),
                            ctx.http.clone(),
                        )),
                    ).await else {
                        return Embed::create_error_respose(username, "error_title", "error_desc");
                    };
                }

                Err(_) => {
                    return Embed::create_error_respose(username, "Cannot join", "Could not join the channel.");
                }
            }
        }
    } else {
        let self_channel = self_channel.expect("Cannot get self channel"); // TODO: match

        // Check if user is in the same VC as the bot
        if self_channel != connect_to {
            return Embed::create_error_respose(
                username,
                "You are not in my VC",
                "You have to be in my VC in order to controll the music.",
            );
        }
    }

    // Get query
    let do_search = !query.starts_with("http");
    let search_query = if do_search {
        match SearchEngines::YouTube.to_query(&query) {
            Ok(x) => x,
            Err(_) => {
                return Embed::create_error_respose(username, "Cannot generate query", "Could not generate a seach query..");
            }
        }
    } else { // Allow piped links
        query.to_string().replace("https://piped.moonleay.net/", "https://youtube.com/")
    };

    // Create source
    let loaded_tracks = llc.load_tracks(guild_id.get(), &search_query).await.expect("Fuck.");

    let mut playlist_info = None;

    let tracks: Vec<TrackInQueue> = match loaded_tracks.data {
        Some(TrackLoadData::Track(x)) => vec![x.into()],
        Some(TrackLoadData::Search(x)) => vec![x[0].clone().into()],
        Some(TrackLoadData::Playlist(x)) => {
            playlist_info = Some(x.info);
            x.tracks.iter().map(|x| x.clone().into()).collect()
        }
        _ => {
            return Embed::create_error_respose(
                username,
                "500: Server failure",
                "Something went wrong when loading the track...");
        }
    };

    let mut response_title = "";
    let mut preview_data = None;

    if let Some(info) = playlist_info {
        response_title = "Added playlist to queue";
        preview_data = Some(Preview {
            title: info.name.to_string(),
            artist: None,
            duration: None,
            thumbnail: None,
            link: None, 
        });
    } else {
        let track = tracks[0].track.clone();

        response_title = "Added to queue";
        preview_data = Some(Preview {
            title: track.info.title,
            artist: Some(track.info.author),
            duration: Some(track.info.length),
            thumbnail: track.info.artwork_url,
            link: track.info.uri
        });
    }

    let player = match llc.get_player_context(guild_id.get()) {
        Some(player) => player,
        None => {
            return Embed::create_error_respose(username, "Cannot get player", "Could not get player context.");
        }
    };

    // let q = player.get_queue();
    // q.append(tracks.into());

    let Ok(_) = player.play_now(&tracks[0].track).await else {
        return Embed::create_error_respose(username, "Error playing", "Could not play track.");
    };

    // if let Ok(player_data) = player.get_player().await {
    //     if player_data.track.is_none() && q.get_track(0).await.is_ok_and(|x| x.is_some()) {
    //         player.skip();
    //     }
    // }


    
    Embed::create_playing(preview_data.unwrap(), username, &response_title).await
}

/// Attempt to skip the song, which is currently playing. Do nothing if there is no next song
pub async fn attempt_to_skip_current_song(
    ctx: &Context,
    llc: &LavalinkClient,
    guild_id: &GuildId,
    user_id: &UserId,
    username: &str,
) -> CreateEmbed {
    if !user_util::is_user_connected_to_vc(ctx, guild_id, user_id).await {
        return Embed::create_error_respose(
            username,
            "You are not connected to a VC",
            "Connect to my vc to start controlling the music.",
        );
    }

    let user_channel = match get_vc_id(ctx, guild_id, user_id).await {
        Some(channel_id) => channel_id,
        None => {
            return Embed::create_error_respose(username, "Error", "Cannot find channel_id.");
        }
    };

    let self_channel = user_util::get_self_vc_id(ctx, guild_id).await;
    if !user_util::is_self_connected_to_vc(ctx, guild_id).await {
        // Self is not connected to vc, cannot skip.

        return Embed::create_error_respose(username, "Not connected", "I am not connected and I am not playing anything, therefore you cannot skip.");
    } else {
        let self_channel = self_channel.expect("Cannot get self channel");

        // Check if user is in the same VC as the bot
        if self_channel != user_channel {
            return Embed::create_error_respose(
                username,
                "You are not in my VC",
                "You have to be in my VC in order to control the music.",
            );
        }
    }

    let Some(player_context) = llc.get_player_context(guild_id.get()) else {
        return Embed::create_error_respose(username, "Not playing", "There is no player context and therefore there is nothing playing.");
    };

    let Ok(player) = player_context.get_player().await else {
        return Embed::create_error_respose(username, "Can't get player", "Cannot get player from player context.");
    };


    if let Some(_) = player.track {
        let Ok(_) = player_context.skip() else {
            return Embed::create_error_respose(username, "Cannot skip", "Could not skip track.");
        };
        let ct = match player_context.get_player().await.expect("Can't get player").track {
            Some(data) => data.info,
            None => { // Disconnect
                let songbird = songbird::get(ctx).await.unwrap().clone();

                let Ok(_) = llc.delete_player(guild_id.get()).await else {
                    return Embed::create_error_respose(username, "Cannot delete player", "Could not delete player.");
                };
                if songbird.get(GuildId::new(guild_id.get())).is_some() {
                    let Ok(_) = songbird.remove(GuildId::new(guild_id.get())).await else {
                        return Embed::create_error_respose(username, "Cannot remove ref", "Cannot remove songbird ref.");
                    };
                }

                return Embed::create_error_respose(username, "No track to skip to", "Therefore I am leaving the channel.")
            }
        };

        let preview_data = Preview {
            title: ct.title,
            artist: Some(ct.author.to_string()),
            duration: Some(ct.length),
            thumbnail: ct.artwork_url,
            link: ct.uri
        };
        
        return Embed::create_playing(preview_data, username, "Skipped; Now playing").await;
    }
    
    return Embed::create_error_respose(username, "Not playing", "I'm currently not playing anything.");


}

/// Try to clear the queue and stop playing. Also leave the vc
pub async fn attempt_to_stop(
    ctx: &Context,
    llc: &LavalinkClient,
    guild_id: &GuildId,
    user_id: &UserId,
    username: &str,
) -> CreateEmbed {
    if !user_util::is_self_connected_to_vc(ctx, guild_id).await {
        // Bot is not connectd to vc; no need to dc
        return Embed::create_error_respose(
            username,
            "Bot is not connected",
            "And therefore there is no need to do anything.",
        );
    }
    let self_channel = user_util::get_self_vc_id(ctx, guild_id)
        .await
        .expect("Cannot get self channel");
    let connect_to = get_vc_id(ctx, guild_id, user_id)
        .await
        .expect("Cannot get channel id");

    // Check if user is in the same VC as the bot
    if self_channel != connect_to {
        return Embed::create_error_respose(
            username,
            "You are not in my VC.",
            "Connect to my VC to controll the music.",
        );
    }

    let player = llc.get_player_context(guild_id.get()).expect("Can't get player context.");
    player.get_queue().clear();

    let stopped = match leave(ctx, llc, guild_id).await {
        Ok(stopped) => stopped,
        Err(e) => {
            println!("Error while stopping: {:?}", e);
            return Embed::create_error_respose(
                username,
                "There was an error",
                "Tell moonleay to check the logs.",
            );
        }
    };

    if !stopped {
        Embed::create_error_respose(
            username,
            "Can't stop, what ain't running",
            "I am not connected.\nI cant stop doing something, when I'm not doing it.",
        )
    } else {
        Embed::create_success_response(username, "I stopped and left", "Just like you girlfriend.")
    }
}

/// Make the bot leave the voice channel. Returns Ok(true) if bot was connected, returns Ok(false) if bot was not connected. Returns Err if something went wrong.
pub async fn leave(ctx: &Context, llc: &LavalinkClient, guild_id: &GuildId) -> Result<bool, JoinError> {
    let Some(songbird) = songbird::get(ctx)
    .await.clone() else {
        return Err(JoinError::NoSender);
    };

    if llc.get_player_context(guild_id.get()).is_some() {
        let Ok(_) = llc.delete_player(guild_id.get()).await else {
            return Err(JoinError::NoCall)
        };
    }

    if songbird.get(GuildId::new(guild_id.get())).is_some() {
        let Ok(_) = songbird.remove(GuildId::new(guild_id.get())).await else {
            return Err(JoinError::Dropped);
        };
    }

    return Ok(true)
}
